# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import testtools

from shaker.engine.executors import netperf


IP = '10.0.0.10'
AGENT = {'slave': {'ip': IP}}


class TestNetperfExecutor(testtools.TestCase):

    def test_get_command(self):
        executor = netperf.NetperfExecutor({}, AGENT)

        expected = {'data': ('netperf -H %s -l 60 -t TCP_STREAM') % IP,
                    'type': 'program'}
        self.assertEqual(expected, executor.get_command())

    def test_get_command_options(self):
        executor = netperf.NetperfExecutor(
            {'method': 'UDP_STREAM', 'time': 30}, AGENT)

        expected = {'data': ('netperf -H %s -l 30 -t UDP_STREAM') % IP,
                    'type': 'program'}
        self.assertEqual(expected, executor.get_command())

    def test_get_command_static_host(self):
        executor = netperf.NetperfExecutor(
            {'method': 'UDP_STREAM', 'host': '10.0.0.20', 'time': 30}, {})

        expected = {'data': 'netperf -H 10.0.0.20 -l 30 -t UDP_STREAM',
                    'type': 'program'}
        self.assertEqual(expected, executor.get_command())



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import testtools

from shaker.agent import agent


class TestAgent(testtools.TestCase):

    @mock.patch('shaker.agent.agent.sleep')
    def test_work_act_idle(self, mock_sleep):
        agent_id = 'the-agent'
        polling_interval = 10
        agent_config = dict(polling_interval=polling_interval)

        mock_socket = mock.Mock()
        mock_socket.recv_json.side_effect = [{}]

        agent.work_act(mock_socket, agent_id, agent_config)

        mock_sleep.assert_called_once_with(polling_interval)
        mock_socket.send_json.assert_called_once_with(
            dict(operation='poll', agent_id=agent_id))
        mock_socket.recv_json.assert_called_once_with()

    @mock.patch('shaker.agent.agent.sleep')
    @mock.patch('shaker.agent.agent.run_command')
    def test_work_act_execute(self, mock_run_command, mock_sleep):
        agent_id = 'the-agent'
        polling_interval = 10
        agent_config = dict(polling_interval=polling_interval)

        mock_socket = mock.Mock()
        mock_socket.recv_json.side_effect = [
            dict(operation='execute'), dict(),
        ]
        execute_result = {'res': 'data'}
        mock_run_command.return_value = execute_result

        agent.work_act(mock_socket, agent_id, agent_config)

        mock_sleep.assert_called_once_with(polling_interval)

        mock_socket.send_json.assert_has_calls([
            mock.call(dict(operation='poll', agent_id=agent_id)),
            mock.call(dict(operation='reply', agent_id=agent_id, res='data')),
        ])

    @mock.patch('shaker.agent.agent.sleep')
    @mock.patch('shaker.agent.agent.time_now')
    @mock.patch('shaker.agent.agent.run_command')
    def test_work_act_schedule(self, mock_run_command, mock_now, mock_sleep):
        agent_id = 'the-agent'
        polling_interval = 10
        start_at = 1234
        now = 1230
        agent_config = dict(polling_interval=polling_interval)

        mock_socket = mock.Mock()
        mock_socket.recv_json.side_effect = [
            dict(operation='execute', start_at=start_at), dict(),
        ]
        execute_result = {'res': 'data'}
        mock_run_command.return_value = execute_result

        mock_now.return_value = now

        agent.work_act(mock_socket, agent_id, agent_config)

        mock_sleep.assert_has_calls([
            mock.call(start_at - now),
            mock.call(polling_interval),
        ])

        mock_socket.send_json.assert_has_calls([
            mock.call(dict(operation='poll', agent_id=agent_id)),
            mock.call(dict(operation='reply', agent_id=agent_id, res='data')),
        ])

    @mock.patch('shaker.agent.agent.sleep')
    @mock.patch('shaker.agent.agent.run_command')
    def test_work_act_configure(self, mock_run_command, mock_sleep):
        agent_id = 'the-agent'
        new_polling_interval = 2
        agent_config = dict(polling_interval=10)

        mock_socket = mock.Mock()
        mock_socket.recv_json.side_effect = [
            dict(operation='configure', polling_interval=new_polling_interval),
            dict(),
        ]

        agent.work_act(mock_socket, agent_id, agent_config)

        mock_sleep.assert_has_calls([
            mock.call(new_polling_interval),
        ])

        mock_socket.send_json.assert_has_calls([
            mock.call(dict(operation='poll', agent_id=agent_id)),
            mock.call(dict(operation='reply', agent_id=agent_id)),
        ])



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import shlex
import tempfile
import time
import uuid

from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log as logging
import sys
import zmq

from shaker.engine import config
from shaker.engine import utils


LOG = logging.getLogger(__name__)


def poll_task(socket, agent_id):
    payload = {
        'operation': 'poll',
        'agent_id': agent_id,
    }
    LOG.debug('Polling task: %s', payload)
    socket.send_json(payload)
    res = socket.recv_json()
    LOG.debug('Received: %s', res)
    return res


def send_reply(socket, agent_id, result):
    message = {
        'operation': 'reply',
        'agent_id': agent_id,
    }
    message.update(result)

    LOG.debug('Sending reply: %s', message)
    socket.send_json(message)
    res = socket.recv_json()
    LOG.debug('Received: %s', res)
    return res


def run_command(command):
    command_stdout, command_stderr = None, None
    start = time.time()

    if command['type'] == 'program':
        command_stdout, command_stderr = processutils.execute(
            *shlex.split(command['data']), check_exit_code=False)

    elif command['type'] == 'script':
        file_name = tempfile.mktemp()
        with open(file_name, mode='w') as fd:
            fd.write(command['data'])
        LOG.debug('Stored script into %s', file_name)
        command_stdout, command_stderr = processutils.execute(
            *shlex.split('bash %s' % file_name), check_exit_code=False)

    else:
        command_stderr = 'Unknown command type : %s' % command['type']

    return dict(stdout=command_stdout, stderr=command_stderr,
                start=start, finish=time.time())


def time_now():
    return time.time()


def sleep(seconds):
    time.sleep(seconds)


def get_socket(endpoint):
    context = zmq.Context()
    socket = context.socket(zmq.REQ)
    socket.connect('tcp://%s' % endpoint)
    return socket


def work_act(socket, agent_id, agent_config):
    task = poll_task(socket, agent_id)
    start_at = task.get('start_at')

    if start_at:
        now = time_now()
        start_at_str = datetime.datetime.fromtimestamp(
            start_at).isoformat()

        if start_at > now:
            LOG.debug('Scheduling task at %s', start_at_str)
            sleep(start_at - now)
        else:
            LOG.warning('Scheduling in the past: %s', start_at_str)

    if task.get('operation') == 'execute':
        result = run_command(task.get('command'))
        send_reply(socket, agent_id, result)
        LOG.info('Finished executing task: %s', task)

    elif task.get('operation') == 'configure':
        if 'polling_interval' in task:
            agent_config['polling_interval'] = task.get('polling_interval')
            send_reply(socket, agent_id, {})
            LOG.info('Agent reconfigured: %s', agent_config)

    sleep(agent_config['polling_interval'])


def work(agent_id, endpoint, polling_interval=config.DEFAULT_POLLING_INTERVAL,
         ignore_sigint=False):
    LOG.info('Agent id is: %s', agent_id)
    LOG.info('Connecting to server: %s', endpoint)

    agent_config = dict(polling_interval=polling_interval)
    LOG.info('Agent config: %s', agent_config)

    socket = get_socket(endpoint)

    while True:
        try:
            work_act(socket, agent_id, agent_config)

        except BaseException as e:
            if isinstance(e, KeyboardInterrupt):
                if ignore_sigint:
                    LOG.info('Got SIGINT, but configured to ignore it')
                else:
                    LOG.info('Process is interrupted')
                    sys.exit(3)
            else:
                LOG.exception(e)
                break


def get_node_uuid():
    s = '%012x' % uuid.getnode()
    return ':'.join([s[i:i + 2] for i in range(0, len(s), 2)])


def main():
    utils.init_config_and_logging(config.COMMON_OPTS + config.AGENT_OPTS)

    endpoint = cfg.CONF.server_endpoint
    polling_interval = cfg.CONF.polling_interval
    agent_id = cfg.CONF.agent_id

    if not agent_id:
        agent_id = get_node_uuid()
        LOG.info('Using node uuid as agent_id: %s', agent_id)

    work(agent_id, endpoint, polling_interval)

if __name__ == "__main__":
    main()



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

import mock
import testtools

from shaker.engine.executors import base as base_executor
from shaker.engine import quorum as quorum_pkg


STEP = 10  # polling interval
LOSS_TIMEOUT = 60
JOIN_TIMEOUT = 600

make_quorum = functools.partial(quorum_pkg.Quorum, polling_interval=STEP,
                                agent_loss_timeout=LOSS_TIMEOUT,
                                agent_join_timeout=JOIN_TIMEOUT)


class DummyExecutor(base_executor.BaseExecutor):
    def __init__(self, duration=STEP):
        super(DummyExecutor, self).__init__({}, None)
        self.duration = duration

    def get_expected_duration(self):
        return self.duration

    def process_reply(self, message):
        return super(DummyExecutor, self).process_reply(message)

    def get_command(self):
        return 'RUN'


class TestQuorum(testtools.TestCase):

    def setUp(self):
        self.mock_time = mock.Mock()
        self._mock_patch = mock.patch('time.time', self.mock_time)
        self._mock_patch.start()
        return super(TestQuorum, self).setUp()

    def tearDown(self):
        self._mock_patch.stop()
        return super(TestQuorum, self).tearDown()

    def _reply(self, expected):
        def reply_handler(reply_message):
            self.assertEqual(expected, reply_message)

        return reply_handler

    def _message_queue_gen(self, event_stream, fail_at_end=True):
        for event in event_stream:
            self.mock_time.return_value = event['time']
            yield (event['msg'], self._reply(event['reply']))

    def test_poll_reply(self):
        self.mock_time.return_value = 0
        event_stream = [
            dict(msg=dict(operation='poll', agent_id='alpha'),
                 reply=dict(operation='execute', command='RUN',
                            start_at=STEP * 2, expected_duration=STEP),
                 time=1),
            dict(msg=dict(operation='reply', agent_id='alpha'),
                 reply=dict(operation='none'),
                 time=20),
        ]

        quorum = make_quorum(self._message_queue_gen(event_stream))
        test_case = {
            'alpha': DummyExecutor()
        }
        result = quorum.execute(test_case)
        self.assertEqual(result.keys(), test_case.keys())

    def test_poll_reply_unknown_agent_ignored(self):
        self.mock_time.return_value = 0
        event_stream = [
            dict(msg=dict(operation='poll', agent_id='alpha'),
                 reply=dict(operation='execute', command='RUN',
                            start_at=STEP * 2, expected_duration=STEP),
                 time=1),
            dict(msg=dict(operation='reply', agent_id='beta'),
                 reply=dict(operation='none'),
                 time=20),
            dict(msg=dict(operation='reply', agent_id='alpha'),
                 reply=dict(operation='none'),
                 time=20),
        ]

        quorum = make_quorum(self._message_queue_gen(event_stream))
        test_case = {
            'alpha': DummyExecutor()
        }
        result = quorum.execute(test_case)
        self.assertEqual(result.keys(), test_case.keys())

    def test_lost_agent(self):
        self.mock_time.return_value = 0
        event_stream = [
            dict(msg=dict(operation='poll', agent_id='alpha'),
                 reply=dict(operation='execute', command='RUN',
                            start_at=STEP * 2, expected_duration=STEP),
                 time=1),
            dict(msg=dict(operation='reply',
                          agent_id=quorum_pkg.HEARTBEAT_AGENT),
                 reply=dict(operation='none'),
                 time=STEP * 10),
        ]

        quorum = make_quorum(self._message_queue_gen(event_stream))
        test_case = {
            'alpha': DummyExecutor()
        }
        result = quorum.execute(test_case)
        self.assertEqual(result.keys(), test_case.keys())
        self.assertEqual('lost', result['alpha']['status'])

    def test_agent_loss_timeout(self):
        """Tests that agent is not marked as lost during loss-timeout."""
        self.mock_time.return_value = 0
        event_stream = [
            dict(msg=dict(operation='poll', agent_id='_lost'),
                 reply=dict(operation='execute', command='RUN',
                            start_at=STEP * 2, expected_duration=STEP),
                 time=1),
            dict(msg=dict(operation='reply',
                          agent_id=quorum_pkg.HEARTBEAT_AGENT),
                 reply=dict(operation='none'),
                 time=LOSS_TIMEOUT),
            dict(msg=dict(operation='reply', agent_id='_lost'),
                 reply=dict(operation='none'),
                 time=LOSS_TIMEOUT),
        ]

        quorum = make_quorum(self._message_queue_gen(event_stream))
        test_case = {
            '_lost': DummyExecutor()
        }
        result = quorum.execute(test_case)
        self.assertEqual(result.keys(), test_case.keys())
        self.assertEqual('ok', result['_lost']['status'])

    def test_good_and_lost(self):
        self.mock_time.return_value = 0
        event_stream = [
            dict(msg=dict(operation='poll', agent_id='_lost'),
                 reply=dict(operation='execute', command='RUN',
                            start_at=STEP * 2, expected_duration=STEP),
                 time=1),
            dict(msg=dict(operation='poll', agent_id='beta'),
                 reply=dict(operation='execute', command='RUN',
                            start_at=STEP * 2, expected_duration=STEP),
                 time=2),
            dict(msg=dict(operation='reply', agent_id='beta'),
                 reply=dict(operation='none'),
                 time=20),
            dict(msg=dict(operation='reply',
                          agent_id=quorum_pkg.HEARTBEAT_AGENT),
                 reply=dict(operation='none'),
                 time=STEP * 10),
        ]

        quorum = make_quorum(self._message_queue_gen(event_stream))
        test_case = {
            '_lost': DummyExecutor(),
            'beta': DummyExecutor(),
        }
        result = quorum.execute(test_case)
        self.assertEqual(set(result.keys()), set(test_case.keys()))
        self.assertEqual('lost', result['_lost']['status'])
        self.assertEqual('ok', result['beta']['status'])

    def test_wait_agentexecutening_long_test(self):
        self.mock_time.return_value = 0
        event_stream = [
            dict(msg=dict(operation='poll', agent_id='alpha'),
                 reply=dict(operation='execute', command='RUN',
                            start_at=STEP * 2, expected_duration=STEP * 9),
                 time=1),
            dict(msg=dict(operation='reply',
                          agent_id=quorum_pkg.HEARTBEAT_AGENT),
                 reply=dict(operation='none'),
                 time=STEP * 4),
            dict(msg=dict(operation='reply', agent_id='alpha'),
                 reply=dict(operation='none'),
                 time=STEP * 9),
        ]

        quorum = make_quorum(self._message_queue_gen(event_stream))
        test_case = {
            'alpha': DummyExecutor(duration=STEP * 9)
        }
        result = quorum.execute(test_case)
        self.assertEqual(result.keys(), test_case.keys())
        self.assertEqual('ok', result['alpha']['status'])

    def test_good_and_interrupted(self):
        self.mock_time.return_value = 0
        event_stream = [
            dict(msg=dict(operation='poll', agent_id='alpha'),
                 reply=dict(operation='execute', command='RUN',
                            start_at=STEP * 2, expected_duration=STEP),
                 time=1),
            dict(msg=dict(operation='poll', agent_id='beta'),
                 reply=dict(operation='execute', command='RUN',
                            start_at=STEP * 2, expected_duration=STEP),
                 time=2),
            dict(msg=dict(operation='reply', agent_id='beta'),
                 reply=dict(operation='none'),
                 time=20),
        ]

        quorum = make_quorum(self._message_queue_gen(event_stream,
                                                     fail_at_end=False))
        test_case = {
            'alpha': DummyExecutor(),
            'beta': DummyExecutor(),
        }
        result = quorum.execute(test_case)
        self.assertEqual(result.keys(), test_case.keys())
        self.assertEqual('interrupted', result['alpha']['status'])
        self.assertEqual('ok', result['beta']['status'])

    def test_join_succeed(self):
        self.mock_time.return_value = 0
        event_stream = [
            dict(msg=dict(operation='poll', agent_id='alpha'),
                 reply=dict(operation='configure', polling_interval=STEP,
                            expected_duration=0),
                 time=STEP),
            dict(msg=dict(operation='reply', agent_id='alpha'),
                 reply=dict(operation='none'),
                 time=STEP * 2),
            dict(msg=dict(operation='poll',
                          agent_id=quorum_pkg.HEARTBEAT_AGENT),
                 reply=dict(operation='none'),
                 time=STEP * 2),
        ]

        quorum = make_quorum(self._message_queue_gen(event_stream))
        result = quorum.join(['alpha'])
        lost = [agent_id for agent_id, r in result.items()
                if r['status'] == 'lost']
        self.assertEqual([], lost)

    def test_join_failed(self):
        self.mock_time.return_value = 0
        event_stream = [
            dict(msg=dict(operation='poll', agent_id='_lost'),
                 reply=dict(operation='configure', polling_interval=STEP,
                            expected_duration=0),
                 time=STEP),
            dict(msg=dict(operation='reply',
                          agent_id=quorum_pkg.HEARTBEAT_AGENT),
                 reply=dict(operation='none'),
                 time=JOIN_TIMEOUT + STEP * 2),
        ]

        quorum = make_quorum(self._message_queue_gen(event_stream))
        result = quorum.join(['_lost'])
        lost = [agent_id for agent_id, r in result.items()
                if r['status'] == 'lost']
        self.assertEqual(['_lost'], lost)



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log as logging


LOG = logging.getLogger(__name__)


class CommandLine(object):
    def __init__(self, command):
        self.tokens = [command]

    def add(self, param_name, param_value=None):
        self.tokens.append('%s' % param_name)
        if param_value is not None:
            self.tokens.append(str(param_value))

    def make(self):
        return dict(type='program', data=' '.join(self.tokens))


class Script(object):
    def __init__(self, script):
        self.script = script

    def make(self):
        return dict(type='script', data=self.script)


class BaseExecutor(object):
    def __init__(self, test_definition, agent):
        super(BaseExecutor, self).__init__()
        self.test_definition = test_definition
        self.agent = agent

    def get_expected_duration(self):
        """Get the expected duration of executor's run
        The value is used by Quorum to calculate expected time of reply.
        :return: time in seconds
        """
        return self.test_definition.get('time') or 60

    def get_command(self):
        return None

    def process_reply(self, message):
        LOG.debug('Test %s on agent %s finished with %s',
                  self.test_definition, self.agent, message)
        return dict(stdout=message.get('stdout'),
                    stderr=message.get('stderr'),
                    start=message.get('start'),
                    finish=message.get('finish'),
                    command=self.get_command())

    def process_failure(self):
        return dict(command=self.get_command())


class ExecutorException(Exception):
    def __init__(self, record, message):
        super(ExecutorException, self).__init__(message)
        self.record = record



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from shaker.engine.executors import base
from shaker.engine import utils


FLENT_EXEC = 'zcat `%s 2>&1 | grep "file written" | grep -Po "\\./\\S+gz"`'
FLENT_EXTRA_TIME = 10  # by default flent adds by 5 secs before and after run


class FlentExecutor(base.BaseExecutor):
    def get_command(self):
        cmd = base.CommandLine('flent')
        cmd.add('-H', self.test_definition.get('host') or
                self.agent['slave']['ip'])
        cmd.add('-l', self.test_definition.get('time') or 60)
        cmd.add('-s', self.test_definition.get('interval') or 1)
        cmd.add(self.test_definition.get('method') or 'tcp_download')
        flent_cmd = cmd.make()
        return base.Script(FLENT_EXEC % flent_cmd['data']).make()

    def get_expected_duration(self):
        return (super(FlentExecutor, self).get_expected_duration() +
                FLENT_EXTRA_TIME)

    def process_reply(self, message):
        result = super(FlentExecutor, self).process_reply(message)

        stdout = result['stdout']
        if not stdout:
            raise base.ExecutorException(
                result,
                'Flent returned no data, stderr: %s' % result['stderr'])

        data = json.loads(stdout)

        series_meta = data['metadata']['SERIES_META']
        columns = sorted(series_meta.keys())
        meta = ([['time', 's']] +
                [[utils.strict(k), series_meta[k]['UNITS']] for k in columns])
        samples = []

        for i in range(len(data['x_values'])):
            line = [data['x_values'][i]]
            for el in columns:
                line.append(data['results'][el][i])
            samples.append(line)

        result['meta'] = meta
        result['samples'] = samples

        return result



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from shaker.engine.executors import flent
from shaker.engine.executors import iperf
from shaker.engine.executors import netperf
from shaker.engine.executors import shell


EXECUTORS = {
    'shell': shell.ShellExecutor,
    'netperf': netperf.NetperfExecutor,
    'iperf': iperf.IperfExecutor,
    'iperf_graph': iperf.IperfGraphExecutor,
    'iperf3': iperf.Iperf3Executor,
    'flent': flent.FlentExecutor,
    '_default': shell.ShellExecutor,
}


def get_executor(test_definition, agent):
    # returns executor of the specified test on the specified agent
    executor_class = test_definition['class']
    klazz = EXECUTORS.get(executor_class, EXECUTORS['_default'])
    return klazz(test_definition, agent)



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import collections
import operator as op
import re


class SLAException(Exception):
    pass


SLAItem = collections.namedtuple('SLAItem', ['record', 'state', 'expression'])
STATE_TRUE = 'OK'
STATE_FALSE = 'FAIL'

# supported operators
operators = {ast.Add: op.add, ast.Sub: op.sub, ast.Mult: op.mul,
             ast.Div: op.truediv, ast.Pow: op.pow, ast.BitXor: op.xor,
             ast.USub: op.neg, ast.Lt: op.lt, ast.Gt: op.gt, ast.LtE: op.le,
             ast.GtE: op.ge, ast.Eq: op.eq, ast.And: op.and_, ast.Or: op.or_,
             ast.Not: op.not_}


def eval_expr(expr, ctx=None):
    """Usage examples:

    >>> eval_expr('2^6')
    4
    >>> eval_expr('2**6')
    64
    >>> eval_expr('1 + 2*3**(4^5) / (6 + -7)')
    -5.0
    >>> eval_expr('11 > a > 5', {'a': 7})
    True
    >>> eval_expr('2 + a.b', {'a': {'b': 2.2}})
    4.2
    """
    ctx = ctx or {}
    return _eval(ast.parse(expr, mode='eval').body, ctx)


def _eval(node, ctx):
    if isinstance(node, ast.Num):
        return node.n
    elif isinstance(node, ast.Name):
        r = ctx.get(node.id)
        if r is None:
            raise SLAException('Value "%s" is not found' % dump_ast_node(node))
        return r
    elif isinstance(node, ast.Str):
        return node.s
    elif isinstance(node, ast.BinOp):
        if isinstance(node.op, ast.RShift):
            return _eval_top(ctx, node)  # the top expression
        elif isinstance(node.op, ast.BitAnd):
            s = _eval(node.left, ctx)
            return ((s is not None) and
                    (re.match(_eval(node.right, ctx), s) is not None))
        else:
            return operators[type(node.op)](_eval(node.left, ctx),
                                            _eval(node.right, ctx))
    elif isinstance(node, ast.UnaryOp):
        return operators[type(node.op)](_eval(node.operand, ctx))
    elif isinstance(node, ast.Compare):
        x = _eval(node.left, ctx)
        r = True
        for i in range(len(node.ops)):
            y = _eval(node.comparators[i], ctx)
            r &= operators[type(node.ops[i])](x, y)
            x = y
        return r
    elif isinstance(node, ast.BoolOp):
        r = _eval(node.values[0], ctx)
        for i in range(1, len(node.values)):
            r = operators[type(node.op)](r, _eval(node.values[i], ctx))
        return r
    elif isinstance(node, ast.Attribute):
        r = _eval(node.value, ctx).get(node.attr)
        if r is None:
            raise SLAException('Value "%s" is not found' % dump_ast_node(node))
        return r
    elif isinstance(node, ast.List):
        records = ctx
        filtered = []
        for record in records:
            for el in node.elts:
                if _eval(el, record):
                    filtered.append(record)
        return filtered
    else:
        raise TypeError(node)


def _eval_top(ctx, node):
    result = []
    # left -- array, right -- condition
    filtered = _eval(node.left, ctx)
    for record in filtered:
        try:
            right = _eval(node.right, record)
            state = (STATE_TRUE if right else STATE_FALSE)
        except (SLAException, TypeError) as e:
            state = str(e)
        result.append(SLAItem(record=record, state=state,
                              expression=dump_ast_node(node.right)))
    return result


def dump_ast_node(node):
    _operators = {ast.Add: '+', ast.Sub: '-', ast.Mult: '*',
                  ast.Div: '/', ast.Pow: '**', ast.BitXor: '^',
                  ast.BitAnd: '&', ast.BitOr: '|', ast.USub: '-',
                  ast.Lt: '<', ast.Gt: '>', ast.LtE: '<=', ast.GtE: '>=',
                  ast.Eq: '==', ast.And: 'and', ast.Or: 'or', ast.Not: 'not'}

    def _format(node):
        if isinstance(node, ast.Num):
            return node.n
        elif isinstance(node, ast.Name):
            return node.id
        elif isinstance(node, ast.Str):
            return '"%s"' % node.s
        elif isinstance(node, ast.BinOp):
            return '%s %s %s' % (_format(node.left), _operators[type(node.op)],
                                 _format(node.right))
        elif isinstance(node, ast.UnaryOp):
            return '%s %s' % (_operators[type(node.op)], _format(node.operand))
        elif isinstance(node, ast.Compare):
            r = '%s' % _format(node.left)
            for i in range(len(node.ops)):
                y = _format(node.comparators[i])
                r = '%s %s %s' % (r, _operators[type(node.ops[i])], y)
            return r
        elif isinstance(node, ast.BoolOp):
            return (' %s ' % _operators[type(node.op)]).join(
                _format(v) for v in node.values)
        elif isinstance(node, ast.Attribute):
            return '%s.%s' % (_format(node.value), node.attr)
        elif isinstance(node, ast.Expression):
            return '(%s)' % _format(node.body)
        else:
            raise TypeError(node)

    return _format(node)



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from shaker.engine import quorum
from shaker.engine import server


class Shaker(object):
    """How to use Shaker as library

>>> from shaker import lib
>>> shaker = lib.Shaker('127.0.0.1:5999')
>>> shaker.run_program('the-agent', 'date')
{
    'status': 'ok',
    'stdout': 'Thu Apr 23 13:09:08 EAT 2015\n',
    'agent': 'the-agent',
    'command': {'data': 'date', 'type': 'program'},
    'stderr': u'',
    'executor': 'shell',
    'test': 'shell',
    'type': 'agent',
    'id': '3a7c3d97-45f1-43ba-8460-2e37e679e3d5'
}
    """
    def __init__(self, server_endpoint, agent_ids=None, polling_interval=1,
                 agent_loss_timeout=60, agent_join_timeout=600):
        self.quorum = quorum.make_quorum(
            agent_ids or [], server_endpoint, polling_interval,
            agent_loss_timeout, agent_join_timeout)

    def _run(self, agent_id, item):
        self.quorum.join({agent_id})
        agents = dict([(agent_id, dict(id=agent_id, mode='alone'))])

        test = {'class': 'shell'}
        test.update(item)

        execution = {'tests': [test]}
        output = dict(records={}, agents={}, tests={})
        server.execute(output, self.quorum, execution, agents)

        return list(output['records'].values())[0]

    def run_program(self, agent_id, program):
        return self._run(agent_id, {'program': program})

    def run_script(self, agent_id, script):
        return self._run(agent_id, {'script': script})



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

try:
    import ConfigParser as configparser
except ImportError:
    import configparser


def split_multiline(value):
    value = [element for element in
             (line.strip() for line in value.split('\n'))
             if element]
    return value


def get_entry_points(config):
    if 'entry_points' not in config:
        return {}
    return dict((option, split_multiline(value))
                for option, value in config['entry_points'].items())


def make(cfg, dest):
    parser = configparser.RawConfigParser()
    parser.read(cfg)
    config = {}
    for section in parser.sections():
        config[section] = dict(parser.items(section))
    entry_points = get_entry_points(config)

    console_scripts = entry_points.get('console_scripts')
    if console_scripts:
        for item in console_scripts:
            tool = item.split('=')[0].strip()
            print('Running %s' % tool)
            os.system('%(tool)s --help > %(dest)s/%(tool)s.txt' %
                      dict(tool=tool, dest=dest))


if len(sys.argv) < 2:
    print('Usage: cli_auto_doc <dest folder>')
    sys.exit(1)


print('Generating docs from help to console tools')
make(cfg='setup.cfg', dest=sys.argv[1])



# Copyright (c) 2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
import re
import os
import sys
import textwrap

import jinja2
import yaml


def make(src, dest):
    scenarios = []
    templates = []

    jinja_env = jinja2.Environment()

    for dirpath, dirnames, filenames in os.walk(src):
        for filename in sorted(filenames):
            if not (filename.endswith('.yaml') or
                    filename.endswith('.hot')):
                continue

            fullpath = os.path.join(dirpath, filename)

            with open(fullpath) as f:
                try:
                    raw = f.read()
                    compiled_template = jinja_env.from_string(raw)
                    rendered_template = compiled_template.render(
                        dict(agents={}))
                    content = yaml.safe_load(rendered_template)

                    s_id = '/'.join(fullpath.split('/')[2:]).split('.')[0]

                    info = dict(title=content.get('title') or s_id,
                                description=content['description'],
                                path=fullpath,
                                scenario_id=s_id)
                    if filename.endswith('.yaml'):
                        scenarios.append(info)
                    else:
                        templates.append(info)
                except Exception as e:
                    print('Failed to read file %s: %s' % (fullpath, e))

    scenarios.sort(key=lambda x: x['scenario_id'])
    templates.sort(key=lambda x: x['scenario_id'])

    with open(dest, 'w') as out:
        print('.. _catalog:\n', file=out)
        print('Scenario Catalog', file=out)
        print('================', file=out)
        print('', file=out)

        print('Scenarios', file=out)
        print('---------', file=out)
        print('', file=out)

        for info in scenarios:
            print_info(out, info, prefix='scenario')
            print('To use this scenario specify parameter ``--scenario %s``.\n'
                  'Scenario source is available at: '
                  'https://github.com/openstack/shaker/blob/master/%s' %
                  (info['scenario_id'], info['path']), file=out)
            print('', file=out)

        print('Heat Templates', file=out)
        print('--------------', file=out)
        print('', file=out)

        for info in templates:
            print_info(out, info, prefix='template')
            print('Template source is available at: '
                  'https://github.com/openstack/shaker/blob/master/%s' %
                  info['path'], file=out)
            print('', file=out)


def print_info(out, info, prefix):
    block_id = re.sub(r'[^\w\d]+', '_', info['title']).lower()
    print('.. _%s_%s:\n' % (prefix, block_id), file=out)
    print(info['title'], file=out)
    print('^' * len(info['title']), file=out)
    print('\n'.join(textwrap.wrap(info['description'], width=79)), file=out)
    print('', file=out)


if len(sys.argv) < 2:
    print('Usage: build_scenario_catalog <dest file>')
    sys.exit(1)


print('Generating scenario catalog')
make(src='shaker/scenarios/', dest=sys.argv[1])



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import functools
import json

import jinja2
from oslo_config import cfg
from oslo_log import log as logging
from subunit import v2 as subunit_v2
import yaml

from shaker.engine import aggregators
from shaker.engine import config
from shaker.engine import sla
from shaker.engine import utils
from shaker.engine import writer


LOG = logging.getLogger(__name__)


def calculate_stats(records, tests):
    # scenario -> test -> concurrency -> [record]
    rec_map = collections.defaultdict(
        functools.partial(collections.defaultdict,
                          functools.partial(collections.defaultdict, list)))

    for record in records.values():
        if 'test' not in record:
            continue

        aggregator = aggregators.get_aggregator(tests[record['test']])
        aggregator.record_summary(record)

        rec_map[record['scenario']][record['test']][
            record['concurrency']].append(record)

    for scenario, per_scenario in rec_map.items():
        for test, per_test in per_scenario.items():
            aggregator = aggregators.get_aggregator(tests[test])

            concurrency_aggregates = []
            for concurrency, per_concurrency in per_test.items():
                summary = aggregator.concurrency_summary(per_concurrency)
                if summary:
                    record_id = utils.make_record_id()
                    summary.update(dict(id=record_id,
                                        scenario=scenario, test=test,
                                        concurrency=concurrency,
                                        type='concurrency'))
                    records[record_id] = summary
                    concurrency_aggregates.append(summary)

            per_test_summary = aggregator.test_summary(concurrency_aggregates)
            if per_test_summary:
                record_id = utils.make_record_id()
                per_test_summary.update(dict(id=record_id,
                                             scenario=scenario, test=test,
                                             type='test'))
                records[record_id] = per_test_summary


def verify_sla(records, tests):
    record_map = collections.defaultdict(list)  # test -> [record]
    for r in records.values():
        if ('test' in r) and ('sla' in tests[r['test']]):
            record_map[r['test']].append(r)

    sla_records = []
    for test_id, records_per_test in record_map.items():
        for sla_expr in tests[test_id]['sla']:
            sla_records += sla.eval_expr(sla_expr, records_per_test)

    return sla_records


def log_sla(sla_records):
    if sla_records:
        LOG.info('*' * 80)
        for item in sla_records:
            test_id = _get_location(item.record) + ':' + item.expression
            LOG.info('%-73s %7s' % (test_id, '[%s]' % item.state))
        LOG.info('*' * 80)


def output_sla(sla_records):
    log_sla(sla_records)
    return [dict(record=item.record['id'], state=item.state,
                 expression=item.expression) for item in sla_records]


def _get_location(record):
    return '.'.join([str(record.get(s))
                     for s in ['scenario', 'test', 'concurrency',
                               'node', 'agent_id'] if record.get(s)])


def save_to_subunit(sla_records, subunit_filename):
    LOG.debug('Writing subunit stream to: %s', subunit_filename)
    fd = None
    state2subunit = {sla.STATE_TRUE: 'success',
                     sla.STATE_FALSE: 'fail'}
    try:
        fd = open(subunit_filename, 'w')
        output = subunit_v2.StreamResultToBytes(fd)

        for item in sla_records:
            output.startTestRun()
            test_id = _get_location(item.record) + ':' + item.expression

            if item.state != sla.STATE_TRUE:
                file_bytes = yaml.safe_dump(
                    item.record, default_flow_style=False).encode('utf8')
                output.status(test_id=test_id, file_name='results',
                              mime_type='text/plain; charset="utf8"', eof=True,
                              file_bytes=file_bytes)

            output.status(test_id=test_id,
                          test_status=state2subunit.get(item.state, 'skip'))
            output.stopTestRun()

        LOG.info('Subunit stream saved to: %s', subunit_filename)
    except IOError as e:
        LOG.error('Error writing subunit stream: %s', e)
    finally:
        if fd:
            fd.close()


def generate_report(data, report_template, report_filename, subunit_filename,
                    book_folder):

    calculate_stats(data['records'], data['tests'])

    sla_records = verify_sla(data['records'], data['tests'])
    data['sla'] = output_sla(sla_records)

    if subunit_filename:
        save_to_subunit(sla_records, subunit_filename)

    # add more filters to jinja
    jinja_env = jinja2.Environment(variable_start_string='[[[',
                                   variable_end_string=']]]',
                                   comment_start_string='[[#',
                                   comment_end_string='#]]')
    jinja_env.filters['json'] = json.dumps
    jinja_env.filters['yaml'] = functools.partial(yaml.safe_dump, indent=2,
                                                  default_flow_style=False)

    alias_mapper = lambda f: config.REPORT_TEMPLATES + '%s.html' % f
    template = utils.read_file(report_template, alias_mapper=alias_mapper)
    compiled_template = jinja_env.from_string(template)
    rendered_template = compiled_template.render(dict(report=data))

    if report_filename:
        LOG.debug('Writing report to: %s', report_filename)
        try:
            utils.write_file(rendered_template, report_filename)
            LOG.info('Report saved to: %s', report_filename)
        except IOError as e:
            LOG.error('Failed to write report file: %s', e)

    if book_folder:
        writer.write_book(book_folder, data)


def main():
    utils.init_config_and_logging(config.REPORT_OPTS + config.INPUT_OPTS)

    outputs = []
    for input_filename in cfg.CONF.input:
        LOG.debug('Reading JSON data from: %s', input_filename)
        outputs.append(json.loads(utils.read_file(input_filename)))

    aggregated = utils.merge_dicts(outputs)
    generate_report(aggregated, cfg.CONF.report_template, cfg.CONF.report,
                    cfg.CONF.subunit, cfg.CONF.book)


if __name__ == "__main__":
    main()



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
import os
import re

import functools
from oslo_config import cfg
from oslo_log import log as logging

from shaker.engine import config
from shaker.engine import deploy
from shaker.engine import executors as executors_classes
from shaker.engine import messaging
from shaker.engine import quorum as quorum_pkg
from shaker.engine import report
from shaker.engine import utils
from shaker.openstack.clients import openstack as openstack_clients


LOG = logging.getLogger(__name__)


def _extend_agents(agents_map):
    extended_agents = {}
    for agent in agents_map.values():
        extended = copy.deepcopy(agent)
        if agent.get('slave_id'):
            extended['slave'] = copy.deepcopy(agents_map[agent['slave_id']])
        if agent.get('master_id'):
            extended['master'] = copy.deepcopy(agents_map[agent['master_id']])
        extended_agents[agent['id']] = extended
    return extended_agents


def _make_test_title(test, params=None):
    s = test.get('title') or test.get('class')
    if params:
        s += ' '.join([','.join(['%s=%s' % (k, v) for k, v in params.items()
                                if k != 'host'])])
    return re.sub(r'[^\x20-\x7e\x80-\xff]+', '_', s)


def _pick_agents(agents, progression):
    # slave agents do not execute any tests
    agents = [a for a in agents.values() if a.get('mode') != 'slave']

    if not progression:
        yield agents
    elif progression in ['arithmetic', 'linear', 'linear_progression']:
        for i in range(len(agents)):
            yield agents[:i + 1]
    elif progression in ['geometric', 'quadratic', 'quadratic_progression']:
        n = len(agents)
        seq = [n]
        while n > 1:
            n //= 2
            seq.append(n)
        seq.reverse()
        for i in seq:
            yield agents[:i]


def run_test(records, quorum, test, agents, progression):

    LOG.debug('Running test %s on all agents', test)
    test_title = test['title']

    for selected_agents in _pick_agents(agents, progression):
        executors = dict((a['id'], executors_classes.get_executor(test, a))
                         for a in selected_agents)

        execution_result = quorum.execute(executors)

        has_interrupted = False
        for agent_id, record in execution_result.items():
            record_id = utils.make_record_id()
            node = agents[agent_id].get('node')
            if node == 'localhost':
                node = test.get('host')

            record.update(dict(
                id=record_id,
                agent=agent_id,
                node=node,
                concurrency=len(selected_agents),
                test=test_title,
                executor=test.get('class'),
                type='agent',
            ))
            records[record_id] = record
            has_interrupted |= record['status'] == 'interrupted'

        if has_interrupted:
            LOG.info('Execution is interrupted')
            return False

    return True


def _pick_tests(tests, matrix):
    matrix = matrix or {}
    for test in tests:
        for params in utils.algebraic_product(**matrix):
            parametrized_test = copy.deepcopy(test)
            parametrized_test.update(params)
            parametrized_test['title'] = _make_test_title(test, params)

            yield parametrized_test


def execute(output, quorum, execution, agents, matrix=None):
    progression = execution.get('progression', execution.get('size'))

    for test in _pick_tests(execution['tests'], matrix):
        output['tests'][test['title']] = test
        proceed = run_test(output['records'],
                           quorum, test, agents, progression)
        if not proceed:
            break  # propagate interruption signal

    LOG.info('Execution is done')


def _under_openstack():
    required = ['os_username', 'os_password', 'os_auth_url']
    for param in required:
        if param not in cfg.CONF:
            return False
        if not cfg.CONF.get(param):
            return False
    return True


def play_scenario(message_queue, scenario):
    deployment = None
    output = dict(scenarios={}, records={}, agents={}, tests={})
    output['scenarios'][scenario['title']] = scenario

    try:
        deployment = deploy.Deployment()

        if _under_openstack():
            openstack_params = utils.pack_openstack_params(cfg.CONF)
            try:
                deployment.connect_to_openstack(
                    openstack_params, cfg.CONF.flavor_name,
                    cfg.CONF.image_name, cfg.CONF.external_net,
                    cfg.CONF.dns_nameservers)
            except openstack_clients.OpenStackClientException:
                raise
            except Exception as e:
                LOG.warning('Failed to connect to OpenStack: %s. Please '
                            'verify parameters: %s', e, openstack_params)
                # try to proceed even if OpenStack connection fails
                # (in case scenario does not need it)

        base_dir = os.path.dirname(scenario['file_name'])
        scenario_deployment = scenario.get('deployment', {})
        server_endpoint = (cfg.CONF.server_endpoint
                           if 'server_endpoint' in cfg.CONF else None)

        agents = deployment.deploy(scenario_deployment, base_dir=base_dir,
                                   server_endpoint=server_endpoint)

        agents = _extend_agents(agents)
        output['agents'] = agents
        LOG.debug('Deployed agents: %s', agents)

        if not agents:
            raise Exception('No agents deployed.')

        if scenario_deployment:
            quorum = quorum_pkg.make_quorum(
                agents.keys(), message_queue,
                cfg.CONF.polling_interval, cfg.CONF.agent_loss_timeout,
                cfg.CONF.agent_join_timeout)
        else:
            # local
            quorum = quorum_pkg.make_local_quorum()

        matrix = cfg.CONF.matrix if 'matrix' in cfg.CONF else None
        if matrix:
            scenario['matrix'] = matrix

        execute(output, quorum, scenario['execution'], agents, matrix)

    except BaseException as e:
        if isinstance(e, KeyboardInterrupt):
            LOG.info('Caught SIGINT. Terminating')
            record = dict(id=utils.make_record_id(), status='interrupted')
        else:
            error_msg = 'Error while executing scenario: %s' % e
            LOG.exception(e)
            record = dict(id=utils.make_record_id(), status='error',
                          stderr=error_msg)
        output['records'][record['id']] = record
    finally:
        if deployment:
            try:
                deployment.cleanup()
            except Exception as e:
                LOG.error('Failed to cleanup the deployment: %s', e,
                          exc_info=True)

    # extend every record with reference to scenario
    for record in output['records'].values():
        record['scenario'] = scenario['title']
    return output


def read_scenario(scenario_name):
    alias_base = scenario_name
    if alias_base[:11] == 'networking/':  # backward compatibility
        LOG.warning('Scenarios from networking/ are moved to openstack/')
        alias_base = 'openstack/' + alias_base[11:]

    alias = '%s%s.yaml' % (config.SCENARIOS, alias_base)
    packaged = utils.resolve_relative_path(alias)

    # use packaged scenario or fallback to full path
    scenario_file_name = packaged or scenario_name
    LOG.debug('Scenario %s is resolved to %s', scenario_name,
              scenario_file_name)

    scenario = utils.read_yaml_file(scenario_file_name)

    schema = utils.read_yaml_file(utils.resolve_relative_path(
        '%s%s.yaml' % (config.SCHEMAS, 'scenario')))
    utils.validate_yaml(scenario, schema)

    scenario['title'] = scenario.get('title') or scenario_file_name
    scenario['file_name'] = scenario_file_name

    return scenario


def act():
    outputs = []

    message_queue = None
    if 'server_endpoint' in cfg.CONF:
        message_queue = messaging.MessageQueue(cfg.CONF.server_endpoint)

    artifacts_dir = cfg.CONF.artifacts_dir
    if artifacts_dir:
        utils.mkdir_tree(artifacts_dir)

    for scenario_name in cfg.CONF.scenario:
        LOG.info('Play scenario: %s', scenario_name)

        scenario = read_scenario(scenario_name)

        play_output = play_scenario(message_queue, scenario)
        outputs.append(copy.deepcopy(play_output))

        # if requested make separate reports
        if artifacts_dir:
            prefix = utils.strict(scenario_name)
            report_name_fn = functools.partial(
                utils.join_folder_prefix_ext, artifacts_dir, prefix)
            utils.write_file(json.dumps(play_output, indent=2),
                             report_name_fn('json'))
            report.generate_report(
                play_output, cfg.CONF.report_template, report_name_fn('html'),
                report_name_fn('subunit'), report_name_fn())

    LOG.info('Generating aggregated report')
    aggregated = utils.merge_dicts(outputs)

    utils.write_file(json.dumps(aggregated, indent=2), cfg.CONF.output)
    LOG.info('Raw output is stored to: %s', cfg.CONF.output)

    report.generate_report(aggregated, cfg.CONF.report_template,
                           cfg.CONF.report, cfg.CONF.subunit, cfg.CONF.book)


def main():
    utils.init_config_and_logging(
        config.COMMON_OPTS + config.OPENSTACK_OPTS + config.SERVER_OPTS +
        config.REPORT_OPTS
    )

    act()


if __name__ == "__main__":
    main()



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

import testtools

from shaker.engine.aggregators import traffic


class TestTrafficAggregator(testtools.TestCase):
    def test_agent_summary(self):
        aggregator = traffic.TrafficAggregator(None)

        original = {
            "stderr": "", "stdout": '',
            "meta": [["time", "s"], ["Ping ICMP", "ms"],
                     ["TCP download", "bit/s"]],
            "samples": [[0, 1.9, None],
                        [1, 2.4, None],
                        [2, 2.6, 60 * 1024 * 1024],
                        [3, 2.2, 65 * 1024 * 1024],
                        [4, 2.2, 61 * 1024 * 1024],
                        [5, 1.9, None]],
        }
        processed = copy.deepcopy(original)
        aggregator.record_summary(processed)

        self.assertFalse('stdout' in processed)

        expected_stats = {
            'Ping ICMP': {
                'max': 2.6,
                'min': 1.9,
                'avg': 2.2,
                'unit': 'ms',
            },
            'TCP download': {
                'max': 65.0,
                'min': 60.0,
                'avg': 62.0,
                'unit': 'Mbit/s',
            }
        }
        self.assertEqual(expected_stats, processed['stats'])

    def test_concurrency_summary(self):
        aggregator = traffic.TrafficAggregator(None)

        original = [
            {
                'agent_id': 'alpha_agent',
                'node': 'alpha',
                'stats': {
                    'Ping ICMP': {
                        'max': 2.6,
                        'min': 1.9,
                        'avg': 2.2,
                        'unit': 'ms',
                    },
                    'TCP download': {
                        'max': 65.0,
                        'min': 60.0,
                        'avg': 62.0,
                        'unit': 'Mbit/s',
                    }
                },
                'chart': [['time', 0, 1, 2, 3, 4, 5],
                          ['Ping ICMP', 1.9, 2.4, 2.6, 2.2, 2.2, 1.9],
                          ['TCP download', None, None, 60.0, 65.0, 61.0,
                          None]]
            },
            {
                'agent_id': 'beta_agent',
                'node': 'beta',
                'stats': {
                    'Ping ICMP': {
                        'max': 3.6,
                        'min': 2.9,
                        'avg': 3.2,
                        'unit': 'ms',
                    },
                    'TCP download': {
                        'max': 75.0,
                        'min': 70.0,
                        'avg': 72.0,
                        'unit': 'Mbit/s',
                    }
                },
                'chart': [['time', 0, 1, 2, 3, 4, 5],
                          ['Ping ICMP', 2.9, 3.4, 3.6, 3.2, 3.2, 2.9],
                          ['TCP download', None, None, 70.0, 75.0, 71.0,
                          None]]
            },
        ]

        aggregate = aggregator.concurrency_summary(original)
        expected_stats = {
            'Ping ICMP': {
                'max': 3.6,
                'min': 1.9,
                'avg': 2.7,
                'unit': 'ms',
            },
            'TCP download': {
                'max': 75.0,
                'min': 60.0,
                'avg': 67.0,
                'unit': 'Mbit/s',
            }
        }

        self.assertEqual(expected_stats, aggregate['stats'])



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from heatclient import exc
from oslo_log import log as logging


LOG = logging.getLogger(__name__)


def create_stack(heat_client, stack_name, template, parameters):
    stack_params = {
        'stack_name': stack_name,
        'template': template,
        'parameters': parameters,
    }

    stack = heat_client.stacks.create(**stack_params)['stack']
    LOG.info('New stack: %s', stack)

    wait_stack_completion(heat_client, stack['id'])

    return stack['id']


def get_stack_status(heat_client, stack_id):
    # stack.get operation may take long time and run out of time. The reason
    # is that it resolves all outputs which is done serially. On the other hand
    # stack status can be retrieved from the list operation. Internally listing
    # supports paging and every request should not take too long.
    for stack in heat_client.stacks.list():
        if stack.id == stack_id:
            return stack.status, stack.stack_status_reason
    raise exc.HTTPNotFound(message='Stack %s is not found' % stack_id)


def wait_stack_completion(heat_client, stack_id):
    reason = None
    status = None

    while True:
        status, reason = get_stack_status(heat_client, stack_id)
        LOG.debug('Stack status: %s', status)
        if status not in ['IN_PROGRESS', '']:
            break

        time.sleep(5)

    if status != 'COMPLETE':
        resources = heat_client.resources.list(stack_id)
        for res in resources:
            if (res.resource_status != 'CREATE_COMPLETE' and
                    res.resource_status_reason):
                LOG.error('Heat stack resource %(res)s of type %(type)s '
                          'failed with %(reason)s',
                          dict(res=res.logical_resource_id,
                               type=res.resource_type,
                               reason=res.resource_status_reason))
        raise Exception('Failed to deploy Heat stack %(id)s. Expected status '
                        'COMPLETE, but got %(status)s. Reason: %(reason)s' %
                        dict(id=stack_id, status=status, reason=reason))


def get_stack_outputs(heat_client, stack_id):
    # try to use optimized way to retrieve outputs, fallback otherwise
    if hasattr(heat_client.stacks, 'output_list'):
        try:
            output_list = heat_client.stacks.output_list(stack_id)['outputs']

            result = {}
            for output in output_list:
                output_key = output['output_key']
                value = heat_client.stacks.output_show(stack_id, output_key)
                result[output_key] = value['output']['output_value']

            return result
        except Exception as e:
            LOG.info('Cannot get output list, fallback to old way: %s', e)

    outputs_list = heat_client.stacks.get(stack_id).to_dict()['outputs']
    return dict((item['output_key'], item['output_value'])
                for item in outputs_list)



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from shaker.engine.executors import base


class NetperfExecutor(base.BaseExecutor):
    def get_command(self):
        cmd = base.CommandLine('netperf')
        cmd.add('-H', self.test_definition.get('host') or
                self.agent['slave']['ip'])
        cmd.add('-l', self.get_expected_duration())
        cmd.add('-t', self.test_definition.get('method') or 'TCP_STREAM')
        return cmd.make()



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import testtools

from shaker.engine import sla


class TestSla(testtools.TestCase):
    def test_eval(self):
        self.assertEqual(2 ** 6, sla.eval_expr('2**6'))
        self.assertTrue(sla.eval_expr('11 > a > 5', {'a': 7}))
        self.assertEqual(42, sla.eval_expr('2 + a.b', {'a': {'b': 40}}))
        self.assertTrue(sla.eval_expr('11 > 7 and 5 < 6'))
        self.assertFalse(sla.eval_expr('(not 11 > 7) or (not 5 < 6)'))

    def test_eval_regex(self):
        self.assertTrue(sla.eval_expr('"some text" & "\w+\s+\w+"'))
        self.assertFalse(sla.eval_expr('"some text" & "\d+"'))

        self.assertFalse(sla.eval_expr('a & "\d+"', {'a': ''}))

    def test_eval_non_existent_ref(self):
        self.assertRaises(sla.SLAException, sla.eval_expr,
                          '2 + a.c', {'a': {'b': 40}})
        self.assertRaises(sla.SLAException, sla.eval_expr,
                          '2 + e.b', {'a': {'b': 40}})

    def test_eval_sla(self):
        records = [{'type': 'agent', 'test': 'iperf_tcp',
                    'stats': {'bandwidth': {'avg': 700}}},
                   {'type': 'agent', 'test': 'iperf_udp',
                    'stats': {'bandwidth': {'avg': 1000}}},
                   {'type': 'node', 'test': 'iperf_tcp',
                    'stats': {'bandwidth': {'avg': 850}}}]

        expr = 'stats.bandwidth.avg > 800'
        sla_records = sla.eval_expr('[type == "agent"] >> (%s)' % expr,
                                    records)
        self.assertEqual([
            sla.SLAItem(record=records[0], state=sla.STATE_FALSE,
                        expression=expr),
            sla.SLAItem(record=records[1], state=sla.STATE_TRUE,
                        expression=expr)],
            sla_records)

        expr = 'stats.bandwidth.avg > 900'
        sla_records = sla.eval_expr('[test == "iperf_udp", type == "node"] >> '
                                    '(%s)' % expr, records)
        self.assertEqual([
            sla.SLAItem(record=records[1], state=sla.STATE_TRUE,
                        expression=expr),
            sla.SLAItem(record=records[2], state=sla.STATE_FALSE,
                        expression=expr)],
            sla_records)

    def test_dump_ast_node(self):
        self.assertEqual('(stderr == "")', sla.dump_ast_node(
            ast.parse('stderr == ""', mode='eval')))

        self.assertEqual('(stderr & ".*")', sla.dump_ast_node(
            ast.parse('stderr & ".*"', mode='eval')))

        self.assertEqual('(stats.bandwidth.avg > 900)', sla.dump_ast_node(
            ast.parse('stats.bandwidth.avg > 900', mode='eval')))

        expr = ('(stats.bandwidth.avg > 900 and not stats.ping.max < 0.5 and '
                'stats.ping.avg < 0.35)')
        self.assertEqual(expr,
                         sla.dump_ast_node(ast.parse(expr, mode='eval')))

    def test_eval_sla_undefined_ref(self):
        records = [{'type': 'agent', 'test': 'iperf_tcp',
                    'stats': {'bandwidth': {'avg': 850}}}]
        expr = 'stats.nonexistent.avg > 800'
        sla_records = sla.eval_expr('[type == "agent"] >> (%s)' % expr,
                                    records)

        self.assertEqual([
            sla.SLAItem(record=records[0],
                        state='Value "stats.nonexistent" is not found',
                        expression=expr)],
            sla_records)



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import re
import time

from novaclient import client as nova_client_pkg
from oslo_log import log as logging


LOG = logging.getLogger(__name__)


class ForbiddenException(nova_client_pkg.exceptions.Forbidden):
    pass


def get_available_compute_nodes(nova_client):
    try:
        return [dict(host=svc.host, zone=svc.zone)
                for svc in nova_client.services.list(binary='nova-compute')
                if svc.state == 'up' and svc.status == 'enabled']
    except nova_client_pkg.exceptions.Forbidden:
        msg = 'Forbidden to get list of compute nodes'
        raise ForbiddenException(msg)


def does_flavor_exist(nova_client, flavor_name):
    for flavor in nova_client.flavors.list():
        if flavor.name == flavor_name:
            return True
    return False


def create_flavor(nova_client, **kwargs):
    try:
        nova_client.flavors.create(**kwargs)
    except nova_client_pkg.exceptions.Forbidden:
        msg = 'Forbidden to create flavor'
        raise ForbiddenException(msg)


def get_server_ip(nova_client, server_name, ip_type):
    server = nova_client.servers.find(name=server_name)
    addresses = server.addresses
    ips = [v['addr'] for v in itertools.chain(*addresses.values())
           if v['OS-EXT-IPS:type'] == ip_type]
    if not ips:
        raise Exception('Could not get IP address of server: %s' % server_name)
    if len(ips) > 1:
        raise Exception('Server %s has more than one IP addresses: %s' %
                        (server_name, ips))
    return ips[0]


def get_server_host_id(nova_client, server_name):
    server = nova_client.servers.find(name=server_name)
    return server.hostId


def check_server_console(nova_client, server_id, len_limit=100):
    console = nova_client.servers.get(server_id).get_console_output(len_limit)

    for line in console.splitlines():
        if (re.search(r'\[critical\]', line, flags=re.IGNORECASE) or
                re.search(r'Cloud-init.*Datasource DataSourceNone\.', line)):
            message = ('Instance %(id)s has critical cloud-init error: '
                       '%(msg)s. Check metadata service availability' %
                       dict(id=server_id, msg=line))
            LOG.error(message)
            return message
        if re.search(r'\[error', line, flags=re.IGNORECASE):
            LOG.error('Error message in instance %(id)s console: %(msg)s',
                      dict(id=server_id, msg=line))
        elif re.search(r'warn', line, flags=re.IGNORECASE):
            LOG.info('Warning message in instance %(id)s console: %(msg)s',
                     dict(id=server_id, msg=line))

    return None


def _poll_for_status(nova_client, server_id, final_ok_states, poll_period=20,
                     status_field="status"):
    LOG.debug('Poll instance %(id)s, waiting for any of statuses %(statuses)s',
              dict(id=server_id, statuses=final_ok_states))
    while True:
        obj = nova_client.servers.get(server_id)

        err_msg = check_server_console(nova_client, server_id)
        if err_msg:
            raise Exception('Critical error in instance %s console: %s' %
                            (server_id, err_msg))

        status = getattr(obj, status_field)
        if status:
            status = status.lower()

        LOG.debug('Instance %(id)s has status %(status)s',
                  dict(id=server_id, status=status))

        if status in final_ok_states:
            break
        elif status == "error" or status == 'paused':
            raise Exception(obj.fault['message'])

        time.sleep(poll_period)


def wait_server_shutdown(nova_client, server_id):
    _poll_for_status(nova_client, server_id, ['shutoff'])


def wait_server_snapshot(nova_client, server_id):
    task_state_field = "OS-EXT-STS:task_state"
    server = nova_client.servers.get(server_id)
    if hasattr(server, task_state_field):
        _poll_for_status(nova_client, server.id, [None, '-', ''],
                         status_field=task_state_field)


def get_flavor(nova_client, flavor_name):
    for flavor in nova_client.flavors.list():
        if flavor.name == flavor_name:
            return flavor
    return None



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import csv
import json
import yaml

from shaker.engine.executors import base
from shaker.engine import utils


def add_common_iperf_params(cmd, executor):
    cmd.add('--client', executor.test_definition.get('host') or
            executor.agent['slave']['ip'])
    cmd.add('--format', 'm')
    if executor.test_definition.get('mss'):
        cmd.add('--mss', executor.test_definition.get('mss'))
    if executor.test_definition.get('buffer_size'):
        cmd.add('--len', executor.test_definition.get('buffer_size'))
    if executor.test_definition.get('udp'):
        cmd.add('--udp')
        if executor.test_definition.get('datagram_size'):
            cmd.add('--len', executor.test_definition.get('datagram_size'))
    if executor.test_definition.get('bandwidth') is not None:
        cmd.add('--bandwidth', executor.test_definition.get('bandwidth'))
    cmd.add('--time', executor.get_expected_duration())
    cmd.add('--parallel', executor.test_definition.get('threads') or 1)
    if executor.test_definition.get('interval'):
        cmd.add('--interval', executor.test_definition.get('interval'))


class IperfExecutor(base.BaseExecutor):
    def get_command(self):
        cmd = base.CommandLine('iperf')
        add_common_iperf_params(cmd, self)
        cmd.add('--nodelay')
        if self.test_definition.get('csv'):
            cmd.add('--reportstyle', 'C')
        return cmd.make()


class IperfGraphExecutor(IperfExecutor):
    def get_command(self):
        self.test_definition['csv'] = True
        self.test_definition['interval'] = 1
        return super(IperfGraphExecutor, self).get_command()

    def process_reply(self, message):
        result = super(IperfGraphExecutor, self).process_reply(message)

        if not result['stdout']:
            raise base.ExecutorException(result, 'Empty result from iperf')

        samples = []
        threads_count = self.test_definition.get('threads') or 1

        for row in csv.reader(result['stdout'].split('\n')):
            if row and len(row) > 8:
                thread = row[5]
                if threads_count > 1 and thread != '-1':
                    # ignore individual per-thread data
                    continue

                start, end = row[6].split('-')
                samples.append([float(end), int(row[8])])

        if samples:
            samples.pop()  # the last line is summary, remove it

        result['samples'] = samples
        result['meta'] = [['time', 's'], ['bandwidth', 'bit/s']]
        return result


class Iperf3Executor(base.BaseExecutor):
    def get_command(self):
        if not self.test_definition.get('interval'):
            self.test_definition['interval'] = 1

        cmd = base.CommandLine('iperf3')
        add_common_iperf_params(cmd, self)
        cmd.add('--json')
        if self.test_definition.get('reverse'):
            cmd.add('--reverse')
        return cmd.make()

    def process_reply(self, message):
        result = super(Iperf3Executor, self).process_reply(message)

        if not result['stdout']:
            raise base.ExecutorException(result, 'Empty result from iperf')

        data = json.loads(result['stdout'])

        # store verbose data in result
        result['verbose'] = yaml.safe_dump(
            dict(start=data['start'], end=data['end']),
            indent=2, default_flow_style=False)

        if 'error' in data:
            raise base.ExecutorException(result, data['error'])

        has_retransmits = False
        if (len(data['intervals']) > 0 and
                'retransmits' in data['intervals'][0]['sum']):
            has_retransmits = True

        if self.test_definition.get('udp'):
            sampler = lambda p: [round(p['end'], 2), p['packets']]
            meta = [['time', 's'], ['packets', 'pps']]
        elif has_retransmits:
            sampler = lambda p: [round(p['end'], 2), p['bits_per_second'],
                                 p['retransmits']]
            meta = [['time', 's'], ['bandwidth', 'bit/s'], ['retransmits', '']]
        else:
            sampler = lambda p: [round(p['end'], 2), p['bits_per_second']]
            meta = [['time', 's'], ['bandwidth', 'bit/s']]

        samples = []
        for point in data['intervals']:
            samples.append(sampler(point['sum']))

        result['samples'] = samples
        result['meta'] = meta

        stats = result['stats'] = {}
        if utils.copy_value_by_path(data, 'end.sum.jitter_ms',
                                    stats, 'jitter.avg'):
            utils.set_value_by_path(stats, 'jitter.unit', 'ms')

        if utils.copy_value_by_path(data, 'end.sum.lost_percent',
                                    stats, 'loss.avg'):
            utils.set_value_by_path(stats, 'loss.unit', '%')

        return result



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import testtools

from shaker.engine.executors import base
from shaker.engine.executors import iperf


IP = '10.0.0.10'
AGENT = {'slave': {'ip': IP}}


class TestIperfGraphExecutor(testtools.TestCase):

    def test_get_command(self):
        executor = iperf.IperfGraphExecutor({}, AGENT)

        expected = {'data': ('iperf --client %s --format m '
                             '--time 60 --parallel 1 --interval 1 --nodelay '
                             '--reportstyle C') % IP,
                    'type': 'program'}
        self.assertEqual(expected, executor.get_command())

    def test_get_command_udp(self):
        executor = iperf.IperfGraphExecutor(
            {'udp': True, 'bandwidth': '100M', 'time': 30,
             'datagram_size': 1470}, AGENT)

        expected = {'data': ('iperf --client %s --format m '
                             '--udp --len 1470 --bandwidth 100M '
                             '--time 30 --parallel 1 --interval 1 --nodelay '
                             '--reportstyle C') % IP,
                    'type': 'program'}
        self.assertEqual(expected, executor.get_command())

    def test_get_command_static_host(self):
        executor = iperf.IperfGraphExecutor({'host': '10.0.0.20'}, {})

        expected = {'data': ('iperf --client %s --format m '
                             '--time 60 --parallel 1 --interval 1 --nodelay '
                             '--reportstyle C') % '10.0.0.20',
                    'type': 'program'}
        self.assertEqual(expected, executor.get_command())

    def test_process_reply(self):
        executor = iperf.IperfGraphExecutor({}, AGENT)
        message = {
            'stdout': """
20150224134955,172.1.7.77,47351,172.1.76.77,5001,3,0.0-1.0,50068684,399507456
20150224134956,172.1.7.77,47351,172.1.76.77,5001,3,1.0-2.0,51605504,412090368
20150224134957,172.1.7.77,47351,172.1.76.77,5001,3,2.0-3.0,50843648,405798912
20150224134957,172.1.7.77,47351,172.1.76.77,5001,3,0.0-3.0,150843648,400000002
"""
        }
        expected = {
            'samples': [
                [1.0, 399507456],
                [2.0, 412090368],
                [3.0, 405798912],
            ],
            'meta': [
                ['time', 's'], ['bandwidth', 'bit/s']
            ]
        }
        reply = executor.process_reply(message)
        self.assertEqual(expected['samples'], reply['samples'],
                         message='Samples data')
        self.assertEqual(expected['meta'], reply['meta'],
                         message='Metadata')

    def test_process_reply_multiple_threads(self):
        executor = iperf.IperfGraphExecutor({'threads': 2}, AGENT)
        message = {
            'stdout': """
20150610102341,10.0.0.3,53479,10.0.0.2,5001,4,0.0-1.0,30277632,242221056
20150610102341,10.0.0.3,53478,10.0.0.2,5001,3,0.0-1.0,23461888,187695104
20150610102341,10.0.0.3,0,10.0.0.2,5001,-1,0.0-1.0,53739520,429916160
20150610102342,10.0.0.3,53479,10.0.0.2,5001,4,1.0-2.0,41418752,331350016
20150610102342,10.0.0.3,53478,10.0.0.2,5001,3,1.0-2.0,22806528,182452224
20150610102342,10.0.0.3,53478,10.0.0.2,5001,3,0.0-2.0,46268416,370147328
20150610102342,10.0.0.3,0,10.0.0.2,5001,-1,1.0-2.0,64225280,513802240
20150610102440,10.0.0.3,53479,10.0.0.2,5001,4,0.0-2.0,71696384,573571072
20150610102440,10.0.0.3,0,10.0.0.2,5001,-1,0.0-2.0,117964800,479810974\n
"""
        }
        expected = {
            'samples': [
                [1.0, 429916160],
                [2.0, 513802240],
            ],
            'meta': [
                ['time', 's'], ['bandwidth', 'bit/s']
            ]
        }
        reply = executor.process_reply(message)
        self.assertEqual(expected['samples'], reply['samples'],
                         message='Samples data')
        self.assertEqual(expected['meta'], reply['meta'],
                         message='Metadata')

    def test_process_empty_reply(self):
        executor = iperf.IperfGraphExecutor({}, AGENT)
        message = {
            'stdout': '',
            'stderr': 'Error!',
        }
        self.assertRaises(
            base.ExecutorException, executor.process_reply, message)


class TestIperf3Executor(testtools.TestCase):

    def test_get_command(self):
        executor = iperf.Iperf3Executor({}, AGENT)

        expected = {'data': ('iperf3 --client %s --format m '
                             '--time 60 --parallel 1 --interval 1 '
                             '--json') % IP,
                    'type': 'program'}
        self.assertEqual(expected, executor.get_command())

    def test_get_command_udp(self):
        executor = iperf.Iperf3Executor(
            {'udp': True, 'bandwidth': '100M', 'time': 30,
             'datagram_size': 1470}, AGENT)

        expected = {'data': ('iperf3 --client %s --format m '
                             '--udp --len 1470 --bandwidth 100M '
                             '--time 30 --parallel 1 --interval 1 '
                             '--json') % IP,
                    'type': 'program'}
        self.assertEqual(expected, executor.get_command())

    def test_get_command_udp_bandwidth_0(self):
        executor = iperf.Iperf3Executor({'udp': True, 'bandwidth': 0}, AGENT)

        expected = {'data': ('iperf3 --client %s --format m '
                             '--udp --bandwidth 0 --time 60 '
                             '--parallel 1 --interval 1 --json') % IP,
                    'type': 'program'}
        self.assertEqual(expected, executor.get_command())

    def test_get_command_reverse(self):
        executor = iperf.Iperf3Executor({'reverse': 'yes'}, AGENT)

        expected = {'data': ('iperf3 --client %s --format m '
                             '--time 60 --parallel 1 --interval 1 '
                             '--json --reverse') % IP,
                    'type': 'program'}
        self.assertEqual(expected, executor.get_command())

    def test_process_reply_tcp(self):
        executor = iperf.Iperf3Executor({}, AGENT)
        message = {
            'stdout': """
{
    "start": {},
    "intervals": [
        {
            "streams": [],
            "sum": {
                "start": 0,
                "end": 1.00002,
                "seconds": 1.00002,
                "bytes": 1712000,
                "bits_per_second": 1.36958e+07,
                "retransmits": 0
            }
        },
        {
            "streams": [],
            "sum": {
                "start": 1.00002,
                "end": 2.00001,
                "seconds": 0.999995,
                "bytes": 2043392,
                "bits_per_second": 1.63472e+07,
                "retransmits": 1
            }
        },
        {
            "streams": [],
            "sum": {
                "start": 2.00001,
                "end": 3.00002,
                "seconds": 1,
                "bytes": 2015872,
                "bits_per_second": 1.61269e+07,
                "retransmits": 0
            }
        }
    ],
    "end": {}
}"""
        }
        expected = {
            'samples': [
                [1.0, 13695800.0, 0],
                [2.0, 16347200.0, 1],
                [3.0, 16126900.0, 0],
            ],
            'meta': [
                ['time', 's'], ['bandwidth', 'bit/s'], ['retransmits', '']
            ]
        }
        reply = executor.process_reply(message)
        self.assertEqual(expected['samples'], reply['samples'],
                         message='Samples data')
        self.assertEqual(expected['meta'], reply['meta'],
                         message='Metadata')

    def test_process_reply_udp(self):
        executor = iperf.Iperf3Executor({'udp': True}, AGENT)
        message = {
            'stdout': """
{
    "start": {},
    "intervals": [
        {
            "streams": [],
            "sum": {
                "start": 0,
                "end": 1.00002,
                "seconds": 1.00002,
                "bytes": 1712000,
                "bits_per_second": 1.36958e+07,
                "packets": 53500
            }
        },
        {
            "streams": [],
            "sum": {
                "start": 1.00002,
                "end": 2.00001,
                "seconds": 0.999995,
                "bytes": 2043392,
                "bits_per_second": 1.63472e+07,
                "packets": 63856
            }
        },
        {
            "streams": [],
            "sum": {
                "start": 2.00001,
                "end": 3.00002,
                "seconds": 1,
                "bytes": 2015872,
                "bits_per_second": 1.61269e+07,
                "packets": 62996
            }
        }
    ],
    "end": {}
}"""
        }
        expected = {
            'samples': [
                [1.0, 53500],
                [2.0, 63856],
                [3.0, 62996],
            ],
            'meta': [
                ['time', 's'], ['packets', 'pps']
            ]
        }
        reply = executor.process_reply(message)
        self.assertEqual(expected['samples'], reply['samples'],
                         message='Samples data')
        self.assertEqual(expected['meta'], reply['meta'],
                         message='Metadata')



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import testtools

from shaker.engine import utils


class TestUtils(testtools.TestCase):
    def setUp(self):
        super(TestUtils, self).setUp()

    def test_split_address_valid(self):
        self.assertEqual(('10.0.0.1', '6777'),
                         utils.split_address('10.0.0.1:6777'))

    def test_split_address_invalid(self):
        self.assertRaises(ValueError, utils.split_address, 'erroneous')

    def test_flatten_dict(self):
        self.assertEqual({}, dict(utils.flatten_dict({})))
        self.assertEqual(
            {'pa_b': 1},
            dict(utils.flatten_dict({'a': {'b': 1}}, prefix='p', sep='_')))
        self.assertEqual(
            {'a': 1, 'b.c': 2, 'b.d': 3},
            dict(utils.flatten_dict({'a': 1, 'b': {'c': 2, 'd': 3}})))

    @mock.patch('os.walk')
    @mock.patch('shaker.engine.utils.resolve_relative_path')
    def test_make_help_options(self, resolve_mock, walk_mock):
        base_dir = 'abc/def'
        abs_dir = '/files/' + base_dir
        walk_mock.side_effect = [
            [(abs_dir, [], ['klm.yaml']), (abs_dir, [], ['ijk.yaml'])],
        ]
        resolve_mock.side_effect = [abs_dir]

        expected = 'List: "ijk", "klm"'
        observed = utils.make_help_options('List: %s', base_dir)
        self.assertEqual(expected, observed)

    @mock.patch('os.walk')
    @mock.patch('shaker.engine.utils.resolve_relative_path')
    def test_make_help_options_subdir(self, resolve_mock, walk_mock):
        base_dir = 'abc/def'
        abs_dir = '/files/' + base_dir
        walk_mock.side_effect = [
            [(abs_dir + '/sub', [], ['klm.yaml']),
             (abs_dir + '/sub', [], ['ijk.yaml'])],
        ]
        resolve_mock.side_effect = [abs_dir]

        expected = 'List: "sub/ijk", "sub/klm"'
        observed = utils.make_help_options('List: %s', base_dir)
        self.assertEqual(expected, observed)

    @mock.patch('os.walk')
    @mock.patch('shaker.engine.utils.resolve_relative_path')
    def test_make_help_options_with_filter(self, resolve_mock, walk_mock):
        base_dir = 'abc/def'
        abs_dir = '/files/' + base_dir
        walk_mock.side_effect = [
            [(abs_dir + '/sub', [], ['klm.yaml']),
             (abs_dir + '/sub', [], ['ijk.html']),
             (abs_dir + '/sub', [], ['mno.yaml'])],
        ]
        resolve_mock.side_effect = [abs_dir]

        expected = 'List: "sub/klm", "sub/mno"'
        observed = utils.make_help_options(
            'List: %s', base_dir, type_filter=lambda x: x.endswith('.yaml'))
        self.assertEqual(expected, observed)

    def test_algebraic_product_empty(self):
        expected = [{}]

        observed = list(utils.algebraic_product())

        self.assertEqual(expected, observed)

    def test_algebraic_product_string(self):
        expected = [{'a': 1, 'b': 'zebra'}, {'a': 2, 'b': 'zebra'}]

        observed = list(utils.algebraic_product(a=[1, 2], b='zebra'))

        self.assertEqual(expected, observed)

    def test_algebraic_product_number(self):
        expected = [{'a': 'x', 'b': 4}, {'a': 2, 'b': 4}]

        observed = list(utils.algebraic_product(a=['x', 2], b=4))

        self.assertEqual(expected, observed)

    def test_strict(self):
        self.assertEqual('some_01_string_a',
                         utils.strict('Some 01-string (brr!) + %% A'))

    def test_set_value_by_path(self):
        data = {}
        utils.set_value_by_path(data, 'jitter.avg', 10)
        self.assertEqual({'jitter': {'avg': 10}}, data)

    def test_set_value_by_path_with_update(self):
        data = {'jitter': {'min': 5}}
        utils.set_value_by_path(data, 'jitter.avg', 10)
        self.assertEqual({'jitter': {'avg': 10, 'min': 5}}, data)

    def test_get_value_by_path(self):
        data = {'jitter': {'min': 5}}
        self.assertEqual(5, utils.get_value_by_path(data, 'jitter.min'))

    def test_get_value_by_path_none(self):
        data = {'jitter': {'min': 5}}
        self.assertIsNone(utils.get_value_by_path(data, 'jitter.avg'))

    def test_copy_value_by_path(self):
        src = {'sum': {'jitter_ms': 7}}
        dst = {}
        res = utils.copy_value_by_path(src, 'sum.jitter_ms', dst, 'jitter.avg')
        self.assertEqual({'jitter': {'avg': 7}}, dst)
        self.assertTrue(res)

    def test_copy_value_by_path_src_not_found(self):
        src = {}
        dst = {}
        res = utils.copy_value_by_path(src, 'sum.jitter_ms', dst, 'jitter.avg')
        self.assertEqual({}, dst)
        self.assertFalse(res)

    def test_merge_dicts(self):
        src = [{'a': {1: 9, 2: 8}, 'b': {1: 3}}, {'a': {3: 7}, 'c': {4: 8}}]
        expected = {'a': {1: 9, 2: 8, 3: 7}, 'b': {1: 3}, 'c': {4: 8}}
        res = utils.merge_dicts(src)
        self.assertEqual(expected, res)



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from shaker.engine.aggregators import base
from shaker.engine.aggregators import traffic


AGGREGATORS = {
    'iperf_graph': traffic.TrafficAggregator,
    'iperf3': traffic.TrafficAggregator,
    'flent': traffic.TrafficAggregator,
    '_default': base.BaseAggregator,
}


def get_aggregator(test_definition):
    # returns executor of the specified test on the specified agent
    executor_class = test_definition['class']
    klazz = AGGREGATORS.get(executor_class, AGGREGATORS['_default'])
    return klazz(test_definition)
